import logging
import sys
import json
import pathlib
import traceback
import csv
import xml.etree.ElementTree


def parse_csv(input_file, source_root):
    """This function parses a csv metrics file into a dictionary object.

    Inputs:
        - input_file: Absolute path to metrics csv file location [string]
        - source_root: Absolute path to source code root directory [string]

    Outputs:
        - metrics_data: Dictionary of metrics values [dict]
    """

    # Initialize variables
    metrics_data = {'project': {}}

    # Read in the metrics data
    with open(input_file) as input_fh:
        raw_metrics_data = list(csv.reader(input_fh, delimiter=','))

        # Get the project level metrics
        for i in range(2, len(raw_metrics_data[0])):
            metric_name = raw_metrics_data[0][i]
            metric_value = raw_metrics_data[1][i]

            # Add the metric as long as it has a value
            if metric_value != 'N/A' and metric_value != '':
                metrics_data['project'][metric_name] = float(metric_value)

        # Get file level metrics
        for i in range(2, len(raw_metrics_data)):
            file_name = str(source_root.joinpath(raw_metrics_data[i][0]))
            metrics_data[file_name] = {}
            for j in range(3, len(raw_metrics_data[0])):
                metric_name = raw_metrics_data[0][j]
                metric_value = raw_metrics_data[i][j]

                # Add the metric as long as it has a value
                if metric_value != 'N/A' and metric_value != '':
                    metrics_data[file_name][metric_name] = float(metric_value)

    return metrics_data


def create_output_file(metrics_data, output_file, tool_metrics_list, file_list):
    """This function generates the output CSV file.

    Inputs:
        - metrics_data: Dictionary containing metrics information [dict]
        - output_file: Absolute path to output file location [string]
        - tool_metrics_list: Dictionary of tool metrics mapped to their common name [dict]
        - file_list: List of all files present in the metrics data [list of strings]

    Outputs:
        None
    """

    # Generate the output file
    with open(output_file, 'w+') as output_fh:
        # Print the header data
        output_fh.write('Scope,Directory,')
        for metric in tool_metrics_list.keys():
            output_fh.write('{},'.format(tool_metrics_list.get(metric)))
        output_fh.write('\n')

        # Print the project level metrics
        output_fh.write('Project,N/A,')
        for metric in tool_metrics_list:
            output_fh.write('{},'.format(metrics_data.get('project_metrics').get(metric, 'N/A')))
        output_fh.write('\n')

        # Print out the file level metrics
        file_list.sort()
        for file_path in file_list:
            output_fh.write('{},{},'.format(file_path, str(pathlib.Path(file_path).parent)))
            for metric in tool_metrics_list.keys():
                output_fh.write('{},'.format(metrics_data.get(file_path).get(metric, 'N/A')))
            output_fh.write('\n')


def parse_codesonar_metrics(raw_analysis_metrics_file, raw_file_metrics_file, parsed_output_file, source_root):
    """This function parses the CodeSonar metrics output file into a more human-readable format.

    Inputs:
        - raw_analysis_metrics_file: Absolute path to the raw analysis-level CodeSonar metrics file [string]
        - raw_file_metrics_file: Absolute path to the raw file-level CodeSonar metrics file [string]
        - parsed_output_file: Absolute path to the parsed output file [string]

    Outputs:
        - codesonar_metrics: Dictionary of all captured metrics [dict]
    """

    # Initialize variables
    cleaned_metrics_data = {'project_metrics': {}}
    metrics_list = {'Top-level file instances': 'Number of Files',
                    'User-defined functions': 'Number of Functions',
                    'Total Lines': 'Total Lines',
                    'Code Lines': 'Lines of Code',
                    'Comment Lines': 'Number of Comments',
                    'Comment Density': 'Comment Density',
                    'Cyclomatic Complexity': 'Cyclomatic Complexity',
                    'Modified Cyclomatic Complexity': 'Modified Cyclomatic Complexity',
                    'Taint Propagator Total': 'Taint Propagator Total',
                    'Taint Sink Total': 'Taint Sink Total',
                    'Taint Source Total': 'Taint Source Total',
                    'Lines with Code': 'Lines with Code',
                    'Lines with Comments': 'Lines with Comments',
                    'Blank Lines': 'Blank Lines',
                    'Mixed Lines': 'In-Line Comments',
                    'Include file instances': 'Number of Includes'}

    # Parse the metrics files if they exist
    if raw_analysis_metrics_file.exists() and raw_file_metrics_file.exists():
        # Read in the analysis-level metrics file
        with open(raw_analysis_metrics_file, 'r') as input_fh:
            analysis_metrics_data = json.load(input_fh)

        # Read in the file-level metrics file
        with open(raw_file_metrics_file, 'r') as input_fh:
            file_metrics_data = json.load(input_fh)

        # Update the data structure
        for metric in analysis_metrics_data.get('metrics'):
            if metric.get('granularity').lower() == 'analysis':
                cleaned_metrics_data['project_metrics'][metric.get('description')] = (metric.get('rows')[0]
                                                                                      .get('metricValue'))

        # Update the data structure
        file_list = []
        file_path = None
        for metric in file_metrics_data.get('metrics'):
            if metric.get('granularity').lower() == 'file':
                for file_metric in metric.get('rows'):
                    # Find the file name in the source tree
                    file_name = file_metric.get('file')
                    file_search = source_root.rglob(file_name)
                    if len(list(file_search)) == 1:
                        file_path = str(next(source_root.rglob(file_name)).relative_to(source_root))
                    else:
                        print('ERROR: Could not resolve file path {}'.format(file_name))

                    # Add the file to the list
                    if file_path not in cleaned_metrics_data.keys():
                        cleaned_metrics_data[file_path] = {}
                        file_list.append(file_path)

                    # Add the metric to the dictionary
                    cleaned_metrics_data[file_path][metric.get('description')] = file_metric.get('metricValue')
                    cleaned_metrics_data[file_path]['Top-level file instances'] = 1

        # Calculate comment density
        for item in cleaned_metrics_data.keys():
            if int(cleaned_metrics_data[item]['Code Lines']) != 0:
                comment_density = round(int(cleaned_metrics_data[item]['Comment Lines']) /
                                        int(cleaned_metrics_data[item]['Code Lines']) * 100, 2)
            else:
                comment_density = 'N/A'
            cleaned_metrics_data[item]['Comment Density'] = comment_density

        # Check to make sure we have data
        if cleaned_metrics_data:
            # Generate the output file
            create_output_file(cleaned_metrics_data, parsed_output_file, metrics_list, file_list)
        else:
            logging.warning('\tCould not parse metrics data. Check log for more information.')
    else:
        logging.warning('\tSome metrics data is missing. Check log for more information.')

    return cleaned_metrics_data


def parse_sonarqube_metrics(metrics_directory, parsed_output_file):
    """This function parses the SonarQube metrics output file into a more human-readable format.

    Inputs:
        - metrics_directory: Absolute path to the directory containing raw SonarQube metrics data [string]
        - parsed_output_file: Absolute path to the parsed output file [string]

    Outputs:
        - cleaned_metrics_data: Dictionary of all captured metrics [dict]
    """

    # Initialize variables
    cleaned_metrics_data = {}
    project_metrics_file = metrics_directory.joinpath('sonarqube_metrics_project.json')
    metrics_files = list(metrics_directory.glob('sonarqube_metrics_file_*.json'))
    metrics_list = {'files': 'Number of Files',
                    'functions': 'Number of Functions',
                    'lines': 'Total Lines',
                    'ncloc': 'Lines of Code',
                    'comment_lines': 'Number of Comments',
                    'comment_density': 'Comment Density',
                    'complexity': 'Cyclomatic Complexity',
                    'cognitive_complexity': 'Cognitive Complexity',
                    'violations': 'Number of Violations',
                    'vulnerabilities': 'Number of Vulnerabilities',
                    'security_hotspots': 'Security Hotspots',
                    'coverage': 'Code Coverage',
                    'line_coverage': 'Code Coverage: Lines',
                    'branch_coverage': 'Code Coverage: Branches',
                    'sqale_index': 'SQALE Index',
                    'duplicated_lines_density': 'Duplication Density'}

    # Read in the metrics data, if it exists
    if project_metrics_file.exists() and metrics_files:
        with open(project_metrics_file, 'r') as input_fh:
            project_metrics_data = json.load(input_fh)

        # Parse the project level metrics
        cleaned_metrics_data = {'project_metrics': {}}
        for project_measures in project_metrics_data.get('baseComponent').get('measures'):
            cleaned_metrics_data['project_metrics'][project_measures.get('metric')] = float(project_measures
                                                                                            .get('value'))

        # Add in the comment density information
        comment_density = round(int(cleaned_metrics_data['project_metrics']['comment_lines']) /
                                int(cleaned_metrics_data['project_metrics']['ncloc']) * 100, 2)
        cleaned_metrics_data['project_metrics']['comment_density'] = comment_density

        # Parse the file level metrics
        file_list = []
        if len(metrics_files) > 0:
            for metrics_file in metrics_files:
                with open(metrics_file, 'r') as input_fh:
                    file_metrics_data = json.load(input_fh)

                # Parse every component
                for source_file in file_metrics_data.get('components'):
                    for file_metric in source_file.get('measures'):
                        if source_file.get('path') not in cleaned_metrics_data.keys():
                            cleaned_metrics_data[source_file.get('path')] = {}
                            file_list.append(source_file.get('path'))
                        cleaned_metrics_data[source_file.get('path')][file_metric.get('metric')] = float(file_metric
                                                                                                         .get('value'))

                    # Add in the comment density information
                    if cleaned_metrics_data[source_file['path']]['ncloc'] > 0:
                        comment_density = round(int(cleaned_metrics_data[source_file['path']]['comment_lines']) /
                                                int(cleaned_metrics_data[source_file['path']]['ncloc']) * 100, 2)
                    else:
                        comment_density = 'N/A'
                    cleaned_metrics_data[source_file.get('path')]['comment_density'] = comment_density

                    # Update complexity measurement if necessary
                    if 'complexity' not in cleaned_metrics_data[source_file.get('path')].keys():
                        cleaned_metrics_data[source_file.get('path')]['complexity'] = 0
                    # Update functions count measurement if necessary
                    if 'functions' not in cleaned_metrics_data[source_file.get('path')].keys():
                        cleaned_metrics_data[source_file.get('path')]['functions'] = 0

        # Generate the output file
        if cleaned_metrics_data:
            create_output_file(cleaned_metrics_data, parsed_output_file, metrics_list, file_list)
        else:
            logging.warning('\tCould parse metrics data. Check log for more details.')
    else:
        logging.warning('\tSome metrics data is missing. Check log for more information.')

    return cleaned_metrics_data


def parse_coverity_metrics(metrics_directory, parsed_output_file):
    """This function parses the SonarQube metrics output file into a more human-readable format.

    Inputs:
        - raw_metrics_file: Absolute path to the directory containing raw Coverity metrics data [string]
        - parsed_output_file: Absolute path to the parsed output file [string]

    Outputs:
        - coverity_metrics: Dictionary of all captured metrics [dict]
    """

    # Initialize variables
    project_metrics_file = metrics_directory.joinpath('output/ANALYSIS.metrics.xml')
    cleaned_metrics_data = {'project_metrics': {}}
    metrics_list = {'files-analyzed': 'Number of Files',
                    'function-metrics-count': 'Number of Functions',
                    'code-lines': 'Lines of Code',
                    'comment-lines': 'Number of Comments'}

    # Read in the project metrics file
    if project_metrics_file.exists():
        project_metrics_tree = xml.etree.ElementTree.parse(project_metrics_file)

        # Parse the project level metrics
        for raw_metric_data in project_metrics_tree.findall('metrics/metric'):
            metric_name = raw_metric_data.find('name').text
            metric_value = raw_metric_data.find('value').text

            # Gather only the metrics of interest
            if metric_name in metrics_list:
                cleaned_metrics_data['project_metrics'][metric_name] = metric_value

        # Generate the output file
        create_output_file(cleaned_metrics_data, parsed_output_file, metrics_list, [])
    else:
        logging.warning('\tSome metrics data is missing. Check log for more information.')

    return cleaned_metrics_data


def parse(metrics_input, output_file, source_dir, tool):
    """This function finds the correct parser for each metrics report.

    Inputs:
        - metrics_input: Absolute path to the raw metrics file or directory containing metrics files [string]
        - output_file: Absolute path to the parsed output file [string]

    Outputs:
        - metrics: Dictionary of all parsed metrics [dict]
    """

    # Initialize variables
    metrics = {}

    # Choose the correct parser
    if tool == 'sonarqube':
        metrics = parse_sonarqube_metrics(metrics_input, output_file)
    elif tool == 'codesonar':
        metrics = parse_codesonar_metrics(metrics_input, output_file, source_dir)
    elif tool == 'coverity':
        metrics = parse_coverity_metrics(metrics_input, output_file)
    else:
        print('ERROR: Unknown metrics file type')
        print(traceback.format_exc())

    return metrics


if __name__ == '__main__':
    parse(pathlib.Path(sys.argv[1]), pathlib.Path(sys.argv[2]), pathlib.Path(sys.argv[3]), sys.argv[4])



import re
import sys
import json
import pathlib
import logging
import traceback
from sarif import loader

WARNING_LINE_REGEX = r'^[a-z]+[0-9]+ <.*>.*:.*:.*:'
CODE_FLOW_REGEX = r'    <.*>.*:.*:.*:'


def create_code_flow(file, line, description):
    """This function creates an internal representation of a code flow to be used by the

    Inputs:
        - file: Absolute path to the source file referenced by the finding [string]
        - line: Line number of the source file being referenced by the findings [int]
        - description: Finding description [list of strings]

    Outputs:
        - code_flow: Dictionary of code flow data [dict]

    """

    code_flow = {'file': file,
                 'line': line,
                 'description': description}

    return code_flow


def create_warning(scrub_id, file, line, description, tool, priority='Low', query='', suppress=False, code_flow=None):
    """This function creates an internal representation of a warning t be used for processing.

    Inputs:
        - id: Finding identifier of the format <tool><count> [string]
        - file: Absolute path to the source file referenced by the finding [string]
        - line: Line number of the source file being referenced by the findings [int]
        - description: Finding description [list of strings]
        - tool: Tool that generated the finding [string]
        - priority: Priority marking for the finding [Low/Med/High]
        - query: Tool query name that generated the finding [string]
        - suppress: Has this finding been suppressed? [bool]
        - code_flows: List of code flows related to the finding [list of dict]

    Outputs:
        - scrub_warning: Dictionary of warning data [dict]
    """

    # Do some type checking
    if code_flow is None:
        code_flow = []
    if type(description) is not list:
        description = [description]

    # Create the warning
    scrub_warning = {'id': scrub_id,
                     'file': file,
                     'line': line,
                     'description': description,
                     'tool': tool,
                     'priority': priority,
                     'query': query,
                     'suppress': suppress,
                     'code_flow': code_flow}

    return scrub_warning


def format_scrub_warning(warning):
    """This function takes an internal representation of a warning and converts it into a SCRUB-formatted string.

    Inputs:
        - warning: Dictionary of finding data [dict]

    Outputs:
        - scrub_warning: SCRUB-formatted warning that can be written to the output file [string]
    """

    # Format the warning
    scrub_warning = (warning.get('id') + ' <' + warning.get('priority') + '> :' + str(warning.get('file')) + ':' +
                     str(warning.get('line')) + ': ' + warning.get('query') + '\n')

    # Add the description
    description = ''
    for line in warning.get('description'):
        description = description + '    ' + line + '\n'

    # Add the code flow
    if len(warning.get('code_flow')) > 0:
        code_flow_description = '    Code flow data:\n'
        for flow_step in warning.get('code_flow'):
            code_flow_description = code_flow_description + '    {}\n    {}:{}\n'.format(flow_step.get('description'),
                                                                                         flow_step.get('file'),
                                                                                         flow_step.get('line'))
    else:
        code_flow_description = ''

    # Add the description
    scrub_warning = scrub_warning + description + code_flow_description + '\n'

    return scrub_warning


def create_scrub_output_file(warnings, output_file):
    """This function writes out raw warnings to a SCRUB formatted output file.

    Inputs:
        - warnings: Dictionary of raw warnings [list of dict]
        - output_file: Absolute path to output file to be created [string]
    """

    # Create the output file
    with open(output_file, 'w', encoding='utf-8') as output_fh:
        # Iterate through every raw warning
        for warning in warnings:
            # Check that the warning isn't suppressed
            if not warning['suppress']:
                # Create the SCRUB formatted warning
                scrub_warning = format_scrub_warning(warning)

                # Write the warning to the output file
                output_fh.write(scrub_warning)


def get_rules_list(warnings):
    """This function gets a list of the rules contained in a set of warnings.

    Inputs:
        - warnings: List of warnings to be examined [list of dicts]

    Outputs:
        - rules_list: List of queries contained in the list of warnings [list of strings]
    """

    # Initialize variables
    rules_list = []

    # Iterate through every warning
    for warning in warnings:
        if warning['query'] not in rules_list and (warning['query'] != ''):
            rules_list.append(warning['query'])

    return rules_list


def parse_scrub(scrub_file, source_root):
    """This function parses a scrub input file and returns a dictionary of findings.

    Inputs:
        - scrub_file: Absolute path to the SCRUB formatted file to bee parsed [string]
        - source_root: Root directory of source code [string]

    Outputs:
        - warning_dict: Dictionary of static analysis findings [dict]
        - query_list: List of queries found in the SCRUB input file [list of strings]
    """

    # Initialize variables
    warning_list = []

    # Import the data
    with open(scrub_file, 'r', encoding='utf-8') as input_fh:
        scrub_data = input_fh.read()

    # Split the warnings
    raw_warnings = list(filter(None, re.split('\n\n', scrub_data)))

    # Find all the warnings in the file
    for raw_warning in raw_warnings:
        warning_lines = list(filter(None, re.split('\n', raw_warning.strip())))

        # Get the location information
        warning_info = list(filter(None, re.split(':', warning_lines[0].strip())))

        # Get the query name if it exists
        if len(warning_info) > 3:
            warning_query = list(warning_lines[0].split(': '))[-1].strip()
        else:
            warning_query = ''

        # Get the warning description
        warning_description = []
        code_flow_data = []
        for i in range(1, len(warning_lines)):
            description_line = warning_lines[i].rstrip().lstrip('    ')

            # Parse code flow data if it exists
            if description_line.lower() == 'code flow data:':
                code_flow_line = i + 1

                # Parse out the code flow if it exists
                for j in range(code_flow_line, len(warning_lines), 2):
                    code_flow_description = warning_lines[j].rstrip().lstrip('    ')
                    code_flow_file = pathlib.Path(warning_lines[j + 1].strip().split(':')[0])
                    code_flow_line = int(warning_lines[j + 1].strip().split(':')[-1])

                    # Generate the code flow object
                    code_flow_data.append(create_code_flow(code_flow_file, code_flow_line, code_flow_description))

                break
            else:
                # Otherwise add the line to the description
                warning_description.append(description_line)

        # Get the values of interest
        warning_id = warning_info[0].split()[0]
        warning_file = pathlib.Path(warning_info[1])
        warning_line = int(warning_info[2])
        warning_tool = re.sub(r'[0-9]', '', warning_id)
        warning_priority = re.sub('[<>]', '', warning_info[0].split()[-1])

        # Update the file path, if necessary
        if warning_file.anchor != '/':
            warning_file = source_root.joinpath(warning_file).resolve()

        # Add the warning to the dictionary
        warning_list.append(create_warning(warning_id, warning_file.resolve(), warning_line, warning_description,
                                           warning_tool, warning_priority, warning_query, code_flow=code_flow_data))

    return warning_list


def format_sarif_for_upload(input_file, output_file, source_root, upload_format):
    """This function pre-processes SARIF files in preparation for import into various tools.

    Inputs:
        - input_file: Absolute path to the SARIF input file to be parsed [string]
        - output_file: Absolute path to output file to be created [string]
        - upload_format: Format of the SARIF output file [string]
    """

    # Initialize variables
    formatted_results = []
    tool_name = input_file.stem

    # Import the SARIF results
    unformatted_results = parse_sarif(input_file, source_root)

    if upload_format == 'sonarqube':
        for warning in unformatted_results:
            if input_file.stem == 'codesonar':
                # warning['description'] = [warning['query'] + " - (" +
                #                           warning['description'][0].split('Server Location: ')[-1] + ")"]
                warning['description'] = [warning['query'] + " - (" + warning['description'][-1] + ")"]
                formatted_results.append(warning)
            elif input_file.stem == 'coverity':
                warning['description'] = [warning['description'][0]]
                formatted_results.append(warning)

        create_sarif_output_file(formatted_results, '2.1.0', output_file, source_root, tool_name)
    elif upload_format == 'codesonar':
        # shutil.copyfile(input_file, output_file)
        for warning in unformatted_results:
            # Add a prefix to the tool name to allow for filtering
            warning['tool'] = 'external-' + warning['tool']
            warning['query'] = warning['tool'].title() + ' ' + warning['query']
            formatted_results.append(warning)
        create_sarif_output_file(formatted_results, '2.1.0', output_file, source_root, tool_name)


def parse_sarif(sarif_filename, source_root):
    """This function parses all the SARIF results into the dictionary list of results.

    NOTE: This function depends on the open-source SARIF parsing library "sarif-tools"
          https://github.com/microsoft/sarif-tools/tree/main

    Inputs:
        - sarif_filename: Absolute path to the SARIF file to be parsed [string]
        - source_root: Absolute path to source root directory [string]

    Outputs:
        - results: List of the dictionary items that represent each filtered analysis result [list of dict]
    """

    # Initialize variables
    results = []
    warning_count = 1

    try:
        # Import the SARIF file data
        sarif_data = loader.load_sarif_file(sarif_filename)

        # Get the tool name
        if sarif_data.get_distinct_tool_names():
            tool_name = sarif_data.get_distinct_tool_names()[0].lower()
        else:
            print("ERROR: No run data found for results file {}".format(sarif_filename))
            raise Exception

        # Update the source root if it can be found in the SARIF data
        if "originalUriBaseIds" in sarif_data.data['runs'][0]:
            # Parse the CodeSonar format
            if 'SRCROOT0' in sarif_data.data['runs'][0]['originalUriBaseIds']:
                source_root = pathlib.Path(sarif_data.data['runs'][0]['originalUriBaseIds']['SRCROOT0']['uri']
                                           .replace('file://', '')).resolve()

        # Iterate through every finding
        for finding in sarif_data.get_results():
            # Set the warning ID
            warning_id = tool_name + str(warning_count).zfill(3)

            # Get the rule ID
            warning_query = finding.get('ruleId')

            # Check if the warning should be suppressed
            if 'suppressions' in finding.keys() and finding.get("suppressions") != []:
                suppress_warning = True
            else:
                suppress_warning = False

            # Get the warning file
            if finding.get('locations'):
                location_data = finding.get('locations')[0].get('physicalLocation')
                warning_file = location_data.get('artifactLocation').get('uri')
                if warning_file.startswith('/'):
                    warning_file = pathlib.Path(warning_file)
                else:
                    warning_file = pathlib.Path(source_root).joinpath(warning_file)

                # Get the line number
                if location_data.get('region'):
                    warning_line = int(location_data.get('region').get('startLine', 0))
                else:
                    warning_line = 0
            else:
                print('WARNING: Location data missing. Could not parse finding {}'.format(warning_query))
                continue

            # Get the warning description
            if finding.get('message').get('text'):
                warning_description = [(finding.get('message').get('text').replace('\n', ''))]
                if finding.get('hostedViewerUri'):
                    # warning_description.append('Server Location: ' + finding.get('hostedViewerUri'))
                    warning_description.append(finding.get('hostedViewerUri'))
            else:
                print('WARNING: Description data missing. Could not parse finding {}'.format(warning_query))
                continue

            # Get any code flow information that exists
            code_flow = []
            if finding.get('codeFlows'):
                for flow in finding.get('codeFlows'):
                    if flow.get('message'):
                        warning_description.append(flow.get('message').get('text'))
                    for thread_location in flow.get('threadFlows')[0].get('locations'):
                        if thread_location.get('location').get('message'):
                            code_flow_file = pathlib.Path(thread_location.get('location')
                                                          .get('physicalLocation').get('artifactLocation').get('uri'))
                            code_flow_line = (thread_location.get('location').get('physicalLocation').get('region')
                                              .get('startLine'))
                            code_flow_description = thread_location.get('location').get('message').get('text')
                            code_flow.append(create_code_flow(code_flow_file, code_flow_line, code_flow_description))

            # Set the ranking
            if 'rank' in finding.keys():
                if int(finding['rank']) > 56:
                    ranking = 'High'
                elif 21 < int(finding['rank']) <= 56:
                    ranking = 'Med'
                else:
                    ranking = 'Low'
            else:
                ranking = 'Low'

            # Add to the warning dictionary
            results.append(create_warning(warning_id, warning_file.resolve(), warning_line, warning_description,
                                          tool_name, ranking, warning_query, suppress_warning, code_flow))

            # Update the warning count
            warning_count = warning_count + 1

    except:      # lgtm [py/catch-base-exception]
        raise Exception

    return results


def create_sarif_output_file(results_list, sarif_version, output_file, source_root, tool_name):
    """This function creates a SARIF formatted output file.

    Inputs:
        - results_list: List of dictionaries representing each warning [list of dicts]
        - sarif_version:
        - output_file:
        - source_root: Absolute path of source root directory [string]
        - tool_name: Name of scanning tool [string]

    Returns:
        - output_file is created at the specified location
    """

    # Initialize variables
    result_item = {}
    rules_list = get_rules_list(results_list)
    sarif_output = {
        'version': sarif_version,
        '$schema': 'https://raw.githubusercontent.com/oasis-tcs/sarif-spec/master/Schemata/sarif-schema-2.1.0.json',
        'runs': [
                   {
                       'tool': {
                           'driver': {
                               'name': tool_name
                           },
                           'rules': []
                       },
                       'results': []
                   }
        ]
    }

    # Iterate through every warning
    for warning in results_list:
        file_index = 0

        # Set the priority level
        result_item['level'] = 'warning'

        # Set the file path
        warning_file = str(warning['file'])

        # Set the rule ID
        result_item['ruleId'] = warning['query']

        if warning.get('description') is not None:
            result_item['message'] = {
                'text': ' '.join(warning['description'])
            }
        if sarif_version == '2.0.0':
            # TODO: CAREFUL USING 2.0.0, STRUCT IS NOT COMPLETELY DEPENDABLE, NEEDS WORK
            sarif_output['runs'][0].update({'resources': rules_list})

            sarif_output['runs'][0].update({'tool': results_list[0]['tool']})
            # TODO: get the logic right for deciding mime type... pass file_list like rules_list
            sarif_output['runs'][0]['files'] = [{
                "mimeType": "text/c",
                "fileLocation": {
                    "uri": results_list[0]['file']
                }
            }]
            result_item['locations'] = [
                {
                    'physicalLocation': {
                        'fileLocation': {
                            # TODO: get the logic right on properly indexing file paths (like with ruleIndex)
                            'fileIndex': file_index
                        },
                        'region': {
                            'startLine': warning['line']
                        }
                    }
                }
            ]
            file_index += 1
        elif sarif_version == '2.1.0':
            # Create the list of sarif rules
            sarif_rules = []
            for rule in rules_list:
                sarif_rules.append({
                    'id': rule,
                    'shortDescription': {
                        'text': rule
                    }
                })
            sarif_output['runs'][0]['tool']['rules'] = sarif_rules
            result_item['locations'] = [{
                'physicalLocation': {
                    'artifactLocation': {
                        'uri': warning_file,
                        'uriBaseId': str(source_root)
                    },
                    'region': {
                        'startLine': warning['line']
                    }
                }
            }]

            if len(warning.get('code_flow')) > 0:
                # Add the codeFlows data to the result
                result_item['codeFlows'] = [{
                                                'message': {
                                                    'text': "Code flow information from {}".format(warning.get('tool'))
                                                },
                                                'threadFlows': [
                                                    {
                                                        'locations': []
                                                    }
                                                ]
                                            }]

                # Get all the code flow data
                for code_flow_item in warning.get('code_flow'):
                    # if str(code_flow_item.get('file')).startswith('/'):
                    if source_root in code_flow_item.get('file').parents:
                        artifact_location = str(code_flow_item.get('file').relative_to(source_root))
                    else:
                        artifact_location = str(code_flow_item.get('file'))

                    location_item = {
                                        'location': {
                                            'message': {
                                                'text': code_flow_item.get('description')
                                            },
                                            'physicalLocation': {
                                                'artifactLocation': {
                                                    'uri': artifact_location
                                                },
                                                'region': {
                                                    'startLine': code_flow_item['line']
                                                }
                                            }
                                        }
                                    }

                    # locations_list.append(location_item)
                    result_item['codeFlows'][0]['threadFlows'][0]['locations'].append(location_item)

        # append fixed warnings to results list and clean dict object
        sarif_output['runs'][0]['results'].append(result_item)
        result_item = {}

    # Create the output file
    with open(output_file, 'w') as output_fh:
        # output_fh.write('{}'.format(json.dumps(sarif_output, indent=4)))
        json.dump(sarif_output, output_fh, indent=4)


def perform_translation(input_file, output_file, source_root, output_format):
    """This function takes in an analysis results file in legacy format (.scrub), then parses and converts the contents
       of each analysis result into the SARIF format.

    Inputs:
        - scrub_filename: The name of the .sarif file to parse and convert. [string]
        - output_filename: The filename to output parsed results to. [string]

    Outputs:
        - custom_exit_code: Exit code that represents whether the module completed with errors.
                            1 - reports any kind of failure with the parser (i.e. read/write, and schema mismatch)
    """

    # Initialize the variables
    exit_code = 1
    parsed_results = []
    tool_name = input_file.stem

    try:
        # Parse the input file
        if input_file.suffix == '.scrub':
            parsed_results = parse_scrub(input_file, source_root)

        elif input_file.suffix == '.sarif':
            parsed_results = parse_sarif(input_file, source_root)

        else:
            # TODO: This should generate an exception
            logging.error('Unknown input file type.')

        # Generate the desired output file
        if output_format == 'scrub':
            create_scrub_output_file(parsed_results, output_file)

        elif output_format.startswith('sarif'):
            # Determine the SARIF version
            sarif_version = output_format.strip('sarifv')

            # Generate the output file
            create_sarif_output_file(parsed_results, sarif_version, output_file, source_root, tool_name)

        else:
            # TODO: This should generate an exception
            logging.error('Unknown output format type.')

        # Update the exit code
        exit_code = 0

    except:     # lgtm [py/catch-base-exception]
        logging.error('Translation could not be performed.')
        logging.error(traceback.format_exc())

        # Update the exit code
        exit_code = 100

    finally:
        return exit_code


if __name__ == '__main__':
    perform_translation(pathlib.Path(sys.argv[1]), pathlib.Path(sys.argv[2]), pathlib.Path(sys.argv[3]), sys.argv[4])



import os
import sys
import argparse
import pathlib
from scrub.utils import scrub_utilities


def parse_arguments():
    """This function handles argument parsing in preparation for diff utility."""

    # Create the parser
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description=diff.__doc__)

    # Add parser arguments
    parser.add_argument('--baseline-scrub', required=True)
    parser.add_argument('--comparison-scrub', required=True)

    # Parse the arguments
    args = vars(parser.parse_args(sys.argv[2:]))

    # Run analysis
    diff(pathlib.Path(args['baseline_scrub']).resolve(), pathlib.Path(args['comparison_scrub']).resolve())


def diff(baseline_scrub_root, comparison_scrub_root):
    """This function diffs two sets of SARIF results directories.

    NOTE: This function depends on the open-source SARIF parsing library "sarif-tools"
          https://github.com/microsoft/sarif-tools/tree/main

    Inputs:
        - baseline_scrub_root: Absolute path to baseline SARIF results directory [string]
        - comparison_scrub_root: Absolute path to comparison SARIF results directory [string]

    Outputs:
        - None
    """

    # Find all the SARIF files in each directory
    baseline_sarif_files = baseline_scrub_root.joinpath('sarif_results').glob('*.sarif')
    comparison_sarif_files = comparison_scrub_root.joinpath('sarif_results').glob('*.sarif')

    for comparison_file in comparison_sarif_files:
        for baseline_file in baseline_sarif_files:
            if comparison_file.stem == baseline_file.stem:
                # Set the output file
                output_file = 'diff_' + comparison_file.name + '.json'

                # Set the command call
                command_call = 'sarif diff -o {} {} {}'.format(output_file, baseline_file, comparison_file)

                # Set the environment for execution
                user_env = os.environ.copy()

                # Execute the analysis
                scrub_utilities.execute_command(command_call, user_env)



import re
import json
import gzip
import pathlib
import xml.etree.ElementTree
from scrub.tools.parsers import translate_results

ID_PREFIX = 'coverity'
warning_count = 1


def parse_cc(raw_input_file, threshold):
    """ This function parses Coverity metrics data to look for functions with high cyclomatic complexity.

    Inputs:
        - raw_input_file: Absolute path to raw Coverity function metrics file [string]
        - threshold: Cyclomatic complexity threshold for identifying violations [int]
        - warning_count: Starting point for warning identifiers [int]

    Outputs:
        - coverity_cc_findings: List of SCRUB formatted findings that violate the CC threshold [list of SCRUB]
    """

    # Initialize variables
    global warning_count
    coverity_cc_findings = []

    # Import the metrics file data
    with gzip.open(raw_input_file, 'rb') as input_fh:
        raw_metrics_data = xml.etree.ElementTree.fromstring('<root>' + str(input_fh.read()) + '</root>')

    # Gather the CC data
    for function_metrics in raw_metrics_data.findall('fnmetric'):
        # Get the file path
        file_path = function_metrics.find('file').text

        # Get the function name
        function_name = function_metrics.find('names').text.split(':')[-1].replace(';', '')

        # Get the metrics data
        function_metrics_data = function_metrics.find('metrics').text.split(';')
        cyclomatic_complexity = int(function_metrics_data[6].split(':')[-1])

        # Check the cyclomatic complexity value
        if cyclomatic_complexity > threshold:
            with open(file_path, 'r') as input_fh:
                for line_number, line in enumerate(input_fh):
                    if re.search(function_name + ".*\(.*\)", line.strip()):
                        warning_id = '%s%03d' % (ID_PREFIX, warning_count)
                        warning_file = pathlib.Path(file_path)
                        ranking = 'LOW'
                        warning_checker = 'SCRUB.HIGH_CC'
                        warning_description = ['High cyclomatic complexity found in function: %s' % function_name,
                                               'Cyclomatic complexity of function is %d, '
                                               'which exceeds the defined threshold of %d. '
                                               'Refactor this function to lower the cyclomatic complexity.' %
                                               (cyclomatic_complexity, threshold)]

                        coverity_cc_findings.append(translate_results.create_warning(warning_id, warning_file,
                                                                                     line_number, warning_description,
                                                                                     'coverity', ranking,
                                                                                     warning_checker))

                        # Increment the warning count
                        warning_count = warning_count + 1
                        break

    return coverity_cc_findings


def parse_json(raw_input_file):
    """This function parses the Coverity internal JSON results format into SCRUB formatted results.

    Inputs:
        - raw_input_file: Absolute path to the file containing raw Coverity warnings [string]

    Outputs:
        - coverity_issues: List of SCRUB formatted Coverity issues [list of SCRUB]
    """

    # Initialize variables
    global warning_count
    coverity_issues = []

    # Read in the input file
    with open(raw_input_file, 'r') as input_fh:
        input_data = json.load(input_fh)

    # Iterate through every issue
    for issue in input_data['issues']:
        # Parse issue data
        warning_id = '%s%03d' % (ID_PREFIX, warning_count)
        warning_file = pathlib.Path(issue['mainEventFilePathname'])
        warning_line = int(issue['mainEventLineNumber'])
        warning_checker = issue['checkerName']
        warning_description = issue['checkerProperties']['subcategoryLongDescription'].encode("unicode_escape").decode("utf-8")
        warning_code_flow = []

        if issue['checkerProperties']['impact'].lower() == 'high':
            ranking = 'High'
        elif issue['checkerProperties']['impact'].lower() == 'medium':
            ranking = 'Med'
        elif issue['checkerProperties']['impact'].lower() == 'low':
            ranking = 'Low'
        elif issue['checkerProperties']['impact'].lower() == 'audit':
            ranking = 'Low'
        else:
            ranking = 'Low'

        # Get the warning description
        for event in issue['events']:
            if event['eventTag'] != 'caretline':
                event_file =event['strippedFilePathname']
                event_line = event['lineNumber']
                event_description = '{}: {}'.format(event['eventTag'],
                                                    event['eventDescription']).encode("unicode_escape").decode("utf-8")

                # Add to the code flow
                warning_code_flow.append(translate_results.create_code_flow(event_file, event_line, event_description))

        coverity_issues.append(translate_results.create_warning(warning_id, warning_file, warning_line,
                                                                warning_description, 'coverity', ranking,
                                                                warning_checker, code_flow=warning_code_flow))

        # Increment the warning count
        warning_count = warning_count + 1

    return coverity_issues


# def parse_warnings(coverity_results_file, coverity_metrics_file, cc_threshold, parsed_output_file):
def parse_warnings(analysis_dir, tool_config_data):
    """This function handles the parsing of Coverity data to generate a SCRUB formatted output file.

    Inputs:
        - coverity_results_file:
        - coverity_metrics_file:
        - cc_threshold:
        - parsed_output_file:
    """

    # Initialize variables
    cc_threshold = int(tool_config_data.get('coverity_cc_threshold'))
    coverity_metrics_file = analysis_dir.joinpath('output/FUNCTION.metrics.xml.gz')
    parsed_output_file = tool_config_data.get('raw_results_dir').joinpath('coverity_raw.scrub')

    # Select the correct parser
    if tool_config_data.get('coverity_json'):
        # Parse the JSON Coverity results
        coverity_findings = parse_json(analysis_dir.joinpath('coverity.json'))
    else:
        # Parse the SARIF Coverity results
        coverity_findings = translate_results.parse_sarif(analysis_dir.joinpath('coverity.sarif'),
                                                          tool_config_data.get('source_dir'))

    # Parse the metrics file, if necessary
    if cc_threshold >= 0:
        coverity_findings = coverity_findings + parse_cc(coverity_metrics_file, cc_threshold)

    # Create the output file
    translate_results.create_scrub_output_file(coverity_findings, parsed_output_file)



import os
import logging
import traceback
from scrub.utils.filtering import create_file_list
from scrub.utils.filtering import filter_results
from scrub.utils import scrub_utilities
from scrub.tools.parsers import translate_results


def initialize_analysis(scrub_conf_data):
    """This function prepares the tool to perform analysis.

    Inputs:
        - scrub_conf_data: Dictionary of scrub.cfg input variables [dict]
    """

    scrub_conf_data.update({'filter_warnings': True})
    scrub_conf_data.update({'filtering_log_file': scrub_conf_data.get('scrub_log_dir').joinpath('filtering.log')})

    return scrub_conf_data


def filter_scrub_results(scrub_conf_data):
    """This function filters the raw SCRUB output files.

    Inputs:
        - scrub_conf_data: Dictionary of SCRUB configuration variables [dict]
    """

    # Create a filtering list
    create_file_list.create_file_list(scrub_conf_data.get('source_dir'),
                                      scrub_conf_data.get('filtering_output_file'),
                                      scrub_conf_data.get('analysis_filters'))

    # Get the list of SCRUB files
    results_files = list(scrub_conf_data.get('raw_results_dir').glob('*.scrub'))

    # Remove files that have already been analyzed
    for results_file in results_files:
        tool_name = results_file.stem.split('_')[0]
        if scrub_conf_data.get('scrub_working_dir').joinpath(tool_name + '.scrub').exists():
            results_files.remove(results_file)

    # Sort the files into groups
    raw_compiler_files = []
    raw_p10_files = []
    raw_generic_files = []
    for results_file in results_files:
        if 'compiler_raw' in results_file.stem:
            raw_compiler_files.append(results_file)
        elif 'p10_raw' in results_file.stem:
            raw_p10_files.append(results_file)
        else:
            raw_generic_files.append(results_file)

    # Filter compiler results
    if raw_compiler_files:
        try:
            # Set the compiler output file path
            filtered_compiler_results_file = scrub_conf_data.get('scrub_analysis_dir').joinpath('compiler.scrub')

            # Parse all the input files
            compiler_results = []
            for results_file in raw_compiler_files:
                # Append the results file
                compiler_results = (compiler_results + translate_results.parse_scrub(results_file,
                                                                                     scrub_conf_data.get('source_dir')))

                # # # Append to the valid warning types
                # valid_warning_types.append(results_file.stem.split('_')[0])

            # Filter the results file
            filter_results.filter_results(compiler_results, filtered_compiler_results_file,
                                          scrub_conf_data.get('filtering_output_file'),
                                          scrub_conf_data.get('query_filters'),
                                          scrub_conf_data.get('source_dir'),
                                          scrub_conf_data.get('enable_micro_filter'),
                                          scrub_conf_data.get('enable_ext_warnings'))

        except:      # lgtm [py/catch-base-exception]
            # Print a status message
            logging.warning("Could not generate output file %s", filtered_compiler_results_file)

            # Print the exception traceback
            logging.debug(traceback.format_exc())

    # Filter P10 results
    if raw_p10_files:
        # Set the output file path
        filtered_p10_results = scrub_conf_data.get('scrub_analysis_dir').joinpath('p10.scrub')

        try:
            # Parse all the input files
            p10_results = []
            for results_file in raw_p10_files:
                # Append the results file
                p10_results = (p10_results + translate_results.parse_scrub(results_file,
                                                                           scrub_conf_data.get('source_dir')))

                # # # Append to the valid warning types
                # valid_warning_types.append(results_file.stem.split('_')[0])

            filter_results.filter_results(p10_results, filtered_p10_results,
                                          scrub_conf_data.get('filtering_output_file'),
                                          scrub_conf_data.get('query_filters'),
                                          scrub_conf_data.get('source_dir'),
                                          scrub_conf_data.get('enable_micro_filter'),
                                          scrub_conf_data.get('enable_ext_warnings'))

        except:     # lgtm [py/catch-base-exception]
            # Print a status message
            logging.warning("Could not generate output file %s", filtered_p10_results)

            # Print the exception traceback
            logging.debug(traceback.format_exc())

    # Filter everything else
    if raw_generic_files:
        for raw_generic_file in raw_generic_files:
            # Get the output file name
            tool_name = raw_generic_file.stem.split('_')[0]
            filtered_generic_results = scrub_conf_data.get('scrub_analysis_dir').joinpath(tool_name + '.scrub')

            try:
                # Import the warning data
                raw_generic_warnings = translate_results.parse_scrub(raw_generic_file,
                                                                     scrub_conf_data.get('source_dir'))

                filter_results.filter_results(raw_generic_warnings, filtered_generic_results,
                                              scrub_conf_data.get('filtering_output_file'),
                                              scrub_conf_data.get('query_filters'),
                                              scrub_conf_data.get('source_dir'),
                                              scrub_conf_data.get('enable_micro_filter'),
                                              scrub_conf_data.get('enable_ext_warnings'))

            except:     # lgtm [py/catch-base-exception]
                # Print a status message
                logging.warning("Could not generate output file %s", filtered_generic_results)

                # Print the exception traceback
                logging.debug(traceback.format_exc())

    # Execute the custom filtering command if it exists
    if scrub_conf_data.get('custom_filter_cmd'):
        scrub_utilities.execute_command(str(scrub_conf_data.get('custom_filter_cmd')), os.environ.copy())


def generate_sarif(scrub_conf_data):
    """This function converts SCRUB formatted output files into SARIF.

    Inputs:
        - scrub_conf_data: Dictionary of SCRUB configuration variables [dict]
    """

    # Find all the SCRUB output files
    scrub_files = scrub_conf_data.get('scrub_analysis_dir').glob('*.scrub')

    for scrub_file in scrub_files:
        # Create the SARIF output file path
        sarif_output_file = scrub_conf_data.get('sarif_results_dir').joinpath(scrub_file.stem + '.sarif')

        # Create a SARIF output file
        translate_results.perform_translation(scrub_file, sarif_output_file, scrub_conf_data.get('source_dir'),
                                              'sarifv2.1.0')


def run_analysis(scrub_conf_data, console_logging=logging.INFO, override=False):
    """This function performs results filtering of raw analysis results.

    Inputs:
        - baseline_conf_data: Dictionary of raw scrub.cfg configuration parameters [dict]
        - console_logging: Level of console logging information to print to console [optional] [enum]
        - override: Force tool execution? [optional] [bool]

    Outputs:
        - log_file/filtering.log: SCRUB log file for the filtering analysis
        - <tool>.scrub: SCRUB-formatted results file after filtering
    """

    # Initialize the analysis
    filtering_conf_data = initialize_analysis(scrub_conf_data)

    # Initialize variables
    filtering_exit_code = 2
    attempt_analysis = filtering_conf_data.get('filter_warnings') or override

    # Perform filtering?
    if attempt_analysis:
        try:
            # Create the logger
            scrub_utilities.create_logger(filtering_conf_data.get('filtering_log_file'), console_logging)

            # Print a status message
            logging.info('')
            logging.info('Perform filtering and distribution...')

            # Remove distributed results
            # TODO: Can this be removed?
            # do_clean.clean_subdirs(scrub_conf_data.get('source_dir'))

            # Filter the results
            filter_scrub_results(scrub_conf_data)

            # Check the status of all the filtered SCRUB output files
            for output_file in scrub_conf_data.get('scrub_analysis_dir').glob('*.scrub'):
                scrub_utilities.check_artifact(output_file, False)

            # Convert the results into SARIF format
            generate_sarif(scrub_conf_data)

            # Check the status of all the filtered SARIF output files
            for output_file in scrub_conf_data.get('scrub_analysis_dir').glob('sarif_results/*.sarif'):
                scrub_utilities.check_artifact(output_file, False)

            # Set the exit code
            filtering_exit_code = 0

        except scrub_utilities.CommandExecutionError:
            # Print a warning message
            logging.warning('Filtering and distribution could not be completed. '
                            'Please see log file %s for more information.',
                            str(filtering_conf_data.get('filtering_log_file')))

            # Print the exception traceback
            logging.warning(traceback.format_exc())

            # Set the exit code
            filtering_exit_code = 100

        except:     # lgtm [py/catch-base-exception]
            # Print a warning message
            logging.warning('Filtering and distribution could not be completed. '
                            'Please see log file %s for more information.',
                            filtering_conf_data.get('filtering_log_file'))

            # Print the exception traceback
            logging.warning(traceback.format_exc())

            # Set the exit code
            filtering_exit_code = 100

        finally:
            # Close the loggers
            logging.getLogger().handlers = []

    # Return the exit code
    return filtering_exit_code



import re
import os
import pathlib
import logging
from scrub.tools.parsers import translate_results

WARNING_LEVEL = 'Low'
ID_PREFIX = 'gbuild'


def get_raw_warning(raw_input_file, initial_index):
    """This function returns the total text of a gbuild warning message.

    Inputs:
        - raw_input_file: Absolute path to the raw gbuild compiler log containing warnings [string]
        - initial_index: The numeric index of where the warnings was found in the log file [int]

    Outputs:
        - warning_text: Full text of the warning starting at initial_index [string]
    """

    # Read in the input file
    with open(raw_input_file, 'r') as input_fh:
        input_data = input_fh.readlines()

    # Initialize variables
    start_point = 0
    end_point = len(input_data)

    # Get the initial line
    current_index = initial_index
    line = input_data[current_index]

    # Get the start point
    while (not line.startswith('\n')) and ('output from compiling' not in line.lower()):
        # Update the start point
        start_point = current_index

        # Update the current line
        current_index = current_index - 1
        line = input_data[current_index]

    # Get the initial line
    current_index = initial_index
    line = input_data[current_index]

    # Get the end point
    while line != '\n':
        # Update the end point
        end_point = current_index

        # Update the current line
        current_index = current_index + 1
        line = input_data[current_index]

    # Store the warning text
    warning_text = []
    for i in range(start_point, end_point):
        warning_text.append(input_data[i])

    return warning_text


def parse_doublecheck_warnings(raw_input_file, parsed_output_file):
    """This function parses a GHS log file for DoubleCheck warnings.

    Inputs:
        - raw_input_file: Absolute path to the raw gbuild compiler log containing warnings [string]
        - parsed_output_file: Absolute path to the desired output file location [string]
    """

    # Initialize the variables
    warning_count = 1

    # Read in the log file contents
    with open(raw_input_file, 'r') as input_fh:
        input_data = input_fh.readlines()

    # Iterate through every line of the input file
    raw_warnings = []
    for i in range(0, len(input_data)):
        # Initialize vriables
        warning_file = None
        warning_line = None
        warning_query = None
        line = input_data[i]

        # Find lines that contain DoubleCheck warnings
        if 'source analysis warning #' in line:
            # Get the raw warning text
            raw_warning = get_raw_warning(raw_input_file, i)

            # Find the file and line number
            for warning_line_itr in raw_warning:
                if "\", line" in warning_line_itr:
                    warning_file = os.path.abspath(list(filter(None, re.split('"', warning_line_itr.strip())))[0])
                    warning_line = int(list(filter(None, re.split(':', re.split('line ',
                                                                                warning_line_itr.strip())[-1])))[0])
                    break

            # Find the query name
            for warning_line_itr in raw_warning:
                if re.search(r'source analysis warning #[0-9]*:', warning_line_itr):
                    warning_query = ('source analysis warning #' +
                                     list(filter(None,
                                                 re.split(':', re.split('#', warning_line_itr.strip())[-1])))[0])
                    break

            # Create the warning description
            warning_message = ['Warning from DoubleCheck:']
            for text in raw_warning:
                warning_message.append(text.rstrip())

            # Set the warning ID
            warning_id = ID_PREFIX + str(warning_count).zfill(3)

            # Add to the warning dictionary
            raw_warnings.append(translate_results.create_warning(warning_id, warning_file, warning_line,
                                                                 warning_message, ID_PREFIX, WARNING_LEVEL,
                                                                 warning_query))

            # Increment the warning count
            warning_count = warning_count + 1

        elif 'source analysis error #' in line:
            # Get the raw warning text
            raw_warning = get_raw_warning(raw_input_file, i)

            # Find the file and line number
            for warning_line_itr in raw_warning:
                if "\", line" in warning_line_itr:
                    warning_file = os.path.abspath(list(filter(None, re.split('"', warning_line_itr.strip())))[0])
                    warning_line = int(list(filter(None, re.split(':', re.split('line ',
                                                                                warning_line_itr.strip())[-1])))[0])
                    break

            # Find the query name
            for warning_line_itr in raw_warning:
                if 'source analysis error #' in warning_line_itr:
                    warning_query = ('source analysis error #' +
                                     list(filter(None, re.split(':', re.split('#', warning_line_itr.strip())[-1])))[0])
                    break

            # Create the warning description
            warning_message = ['Warning from DoubleCheck:']
            for text in raw_warning:
                warning_message.append(text.rstrip())

            # Set the warning ID
            warning_id = ID_PREFIX + str(warning_count).zfill(3)

            # Add to the warning dictionary
            raw_warnings.append(translate_results.create_warning(warning_id, warning_file, warning_line,
                                                                 warning_message, ID_PREFIX, WARNING_LEVEL,
                                                                 warning_query))

            # Increment the warning count
            warning_count = warning_count + 1

        elif ': warning #' in line:
            # Get the raw warning text
            raw_warning = get_raw_warning(raw_input_file, i)

            # Get the warning location data
            warning_location = raw_warning[0].split(':')[0]
            warning_file = warning_location.split(',')[0].replace('"', '').strip()
            warning_line = int(warning_location.split('line ')[-1].strip())
            warning_query = raw_warning[0].split(':')[1].strip()

            # Create the warning description
            warning_message = ['Warning from gbuild:']
            for text in raw_warning:
                warning_message.append(text.rstrip())

            # Set the warning ID
            warning_id = ID_PREFIX + str(warning_count).zfill(3)

            # Add to the warning dictionary
            raw_warnings.append(translate_results.create_warning(warning_id, warning_file, warning_line,
                                                                 warning_message, ID_PREFIX, WARNING_LEVEL,
                                                                 warning_query))

            # Increment the warning count
            warning_count = warning_count + 1

    # Create the SCRUB output file
    translate_results.create_scrub_output_file(raw_warnings, parsed_output_file)


def parse_warnings(analysis_dir, tool_config_data, raw_input_file=None, parsed_output_file=None):
    """This function parses the raw gbuild compiler warnings into the SCRUB format.

    Inputs:
        - analysis_dir: Absolute path to the raw gbuild output file directory [string]
        - tool_config_data: Dictionary of scrub configuration data [dict]
        - raw_input_file: Absolute path to the raw input file [string] [optional]
        - parsed_output_file: Absolute path to the raw output file [string] [optional]
    """

    # Set input file
    if raw_input_file is None:
        raw_input_file = analysis_dir.joinpath('gbuild_build.log')

    # Set the output file
    if parsed_output_file is None:
        parsed_output_file = tool_config_data.get('raw_results_dir').joinpath('gbuild_compiler_raw.scrub')

    # Print a status message
    logging.info('')
    logging.info('\tParsing results...')
    logging.info('\t>> Executing command: get_gbuild_warnings.parse_warnings(%s, %s)', raw_input_file,
                 parsed_output_file)
    logging.info('\t>> From directory: %s', str(pathlib.Path().absolute()))

    # Search for DoubleCheck warnings in the log file
    parse_doublecheck_warnings(raw_input_file, parsed_output_file)



import pathlib
import logging
import traceback
from scrub.utils import scrub_utilities
from scrub.tools.parsers import translate_results


def distribute_warnings(warning_file, source_dir):
    """This function moves warnings to be co-located with the source file of interest.

    Inputs:
        - warning_file: Full path to the file containing SCRUB-formatted warnings [string]
        - source_dir: Full path to the top-level directory of the source code [string]

    Outputs:
        - A series of .scrub directories and output files will be created as necessary
    """

    # Initialize the variables
    warning_type = warning_file.stem
    distributed_files_list = []

    # Print a status message
    logging.info('')
    logging.info('\tMoving results...')
    logging.info('\t>> Executing command: do_gui.distribute_warnings(%s, %s)', str(warning_file), str(source_dir))
    logging.info('\t>> From directory: %s', str(pathlib.Path().absolute()))

    # Update the source root to make it absolute
    source_dir = source_dir.resolve()

    # Parse all the findings from the warning file
    warnings = translate_results.parse_scrub(warning_file, source_dir)

    # Iterate through everything warning
    for warning in warnings:
        # Make sure the warning file is within the source root, but not at the source root
        if warning['file'].exists() and (source_dir != warning['file'].parent):
            # Get the warning directory
            warning_directory = warning['file'].parent

            # Create the scrub output paths
            local_scrub_directory = warning_directory.joinpath('.scrub')
            local_scrub_warning_file = local_scrub_directory.joinpath(warning_type + '.scrub')

            # Create a .scrub directory if it doesn't already exist
            if not local_scrub_directory.exists():
                local_scrub_directory.mkdir()
                local_scrub_directory.chmod(0o755)

            # Write the warning to the output file
            with open(local_scrub_warning_file, 'a') as output_fh:
                output_fh.write('%s' % translate_results.format_scrub_warning(warning))

            # Add the file to the list if it hasn't been added already
            if local_scrub_warning_file not in distributed_files_list:
                distributed_files_list.append(local_scrub_warning_file)

            # Change the permissions of the output file
            local_scrub_warning_file.chmod(0o644)

    return distributed_files_list


def initialize_analysis(tool_conf_data):
    """The purpose of this function is to prepare the tool to perform analysis.

    Inputs:
        - tool_conf_data: Dictionary of scrub.cfg input variables [dict]
    """

    # Initialize the derived variables
    gui_log_file = tool_conf_data.get('scrub_log_dir').joinpath('gui.log')

    # Add derived values to the dictionary
    tool_conf_data.update({'gui_log_file': gui_log_file})


def run_analysis(baseline_conf_data, console_logging=logging.INFO, override=False):
    """This function runs this module based on input from the scrub.cfg file

    Inputs:
        - baseline_conf_data: Dictionary of raw scrub.cfg configuration parameters [dict]
        - console_logging: Level of console logging information to print to console [optional] [enum]
        - override: Force tool execution? [optional] [bool]
    """

    # Import the config file data
    tool_conf_data = baseline_conf_data.copy()
    initialize_analysis(tool_conf_data)

    # Initialize variables
    gui_exit_code = 2
    attempt_analysis = tool_conf_data.get('scrub_gui_export') or override

    # Determine if GUI export can be run
    if attempt_analysis:
        try:
            # Create the logging file, if it doesn't already exist
            scrub_utilities.create_logger(tool_conf_data.get('gui_log_file'), console_logging)

            # Print a status message
            logging.info('')
            logging.info('Perform GUI export...')

            # Get a list of the filtered SCRUB output files
            filtered_output_files = tool_conf_data.get('scrub_analysis_dir').glob('*.scrub')

            # Move the warnings to the appropriate directories
            for filtered_output_file in filtered_output_files:
                distributed_files = distribute_warnings(filtered_output_file, tool_conf_data.get('source_dir'))

                # Check each of the generated files for formatting
                for distributed_file in distributed_files:
                    if len(translate_results.parse_scrub(distributed_file, tool_conf_data.get('source_dir'))) == 0:
                        raise AssertionError

            # Set the exit code
            gui_exit_code = 0

        except:     # lgtm [py/catch-base-exception]
            # Print a warning message
            logging.warning('GUI export could not be performed. Please see log file %s for more information.',
                            str(tool_conf_data.get('gui_log_file')))

            # Print the exception traceback
            logging.warning(traceback.format_exc())

            # Set the exit code
            gui_exit_code = 1

        finally:
            # Close the loggers
            logging.getLogger().handlers = []

            # Update the permissions of the log file if it exists
            if tool_conf_data.get('gui_log_file').exists():
                tool_conf_data.get('gui_log_file').chmod(0o644)

    # Return the exit code
    return gui_exit_code
